/*****************************************************************************
 * Copyright (C) 2014-2024
 *   
 * Michiel Broek <mbroek at mbse dot eu>
 *
 * This file is part of the mbsePi-apps
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * mbsePi-apps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with ThermFerm; see the file COPYING.  If not, write to the Free
 * Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *****************************************************************************/

#include "rdconfig.h"
#include "thermferm.h"
#include "pid.h"
#include "futil.h"
#include "xutil.h"

int			debug = FALSE;
int			foreground = FALSE;
sys_config		Config;			/* System configuration		*/
extern pthread_mutex_t  mutexes[5];

#define MY_ENCODING "utf-8"

const char      UNITMODE[5][8]	= { "OFF", "NONE", "FRIDGE", "BEER", "PROFILE" };
const char	UNITSTAGE[4][12] = { "PRIMARY", "SECONDARY", "TERTIARY", "CARBONATION" };
const char	PROFSTATE[5][6]	= { "OFF", "PAUSE", "RUN", "DONE", "ABORT" };
const char	DEVTYPE[9][6] = { "NA", "W1", "GPIO", "RC433", "DHT", "I2C", "SPI", "SIM", "DOMO" };
const char	DEVPRESENT[4][6] = { "UNDEF", "NO", "YES", "ERROR" };
const char	DEVDIR[7][11] = { "UNDEF", "IN_BIN", "OUT_BIN", "IN_ANALOG", "OUT_ANALOG", "OUT_PWM", "INTERN" };
const char	PIDMODE[3][5] = { "NONE", "AUTO", "BOO" };



/*
 * https://stackoverflow.com/questions/56835876/number-of-decimals-of-a-double-in-json-c-json-h
 */
struct json_object* json_object_new_double_fmt(double d, const char *fmt)
{
    char tmp[128];

    snprintf(tmp, 128, fmt, d);
    return(json_object_new_double_s(d, tmp));
}


void killconfig(void)
{
    units_list		*tmp2, *oldtmp2;
    prof_step		*step, *oldstep;
    devices_list	*device, *olddev;
#ifdef USE_SIMULATOR
    simulator_list	*simulator, *oldsim;
#endif

    if (Config.name)
	free(Config.name);
    Config.name = NULL;

    Config.server_port = 6554;
    Config.websocket_port = 8020;
    if (Config.temp_uuid)
	free(Config.temp_uuid);
    if (Config.hum_uuid)
	free(Config.hum_uuid);
    Config.temp_hum_idx = 0;
    Config.temp_uuid = Config.hum_uuid = NULL;
    Config.temp_value = 20000;
    Config.temp_state = Config.hum_state = 1;	// missing
    Config.hum_value = 50000;
    if (Config.mqtt_host)
	free(Config.mqtt_host);
    Config.mqtt_host = xstrcpy((char *)"localhost");
    Config.mqtt_port = 1883;
    if (Config.mqtt_username)
	free(Config.mqtt_username);
    Config.mqtt_username = NULL;
    if (Config.mqtt_password)
	free(Config.mqtt_password);
    Config.mqtt_password = NULL;
    if (Config.uuid)
	free(Config.uuid);
    Config.uuid = NULL;
 
    for (tmp2 = Config.units; tmp2; tmp2 = oldtmp2) {
	oldtmp2 = tmp2->next;
	if (tmp2->uuid)
	    free(tmp2->uuid);
	if (tmp2->product_uuid)
	    free(tmp2->product_uuid);
	if (tmp2->product_code)
	    free(tmp2->product_code);
	if (tmp2->product_name)
	    free(tmp2->product_name);
	if (tmp2->air_address)
	    free(tmp2->air_address);
	if (tmp2->beer_address)
	    free(tmp2->beer_address);
	if (tmp2->beer_address2)
	    free(tmp2->beer_address2);
	if (tmp2->chiller_address)
	    free(tmp2->chiller_address);
	if (tmp2->heater_address)
	    free(tmp2->heater_address);
	if (tmp2->cooler_address)
	    free(tmp2->cooler_address);
	if (tmp2->fan_address)
	    free(tmp2->fan_address);
	if (tmp2->light_address)
	    free(tmp2->light_address);
	if (tmp2->door_address)
	    free(tmp2->door_address);
	if (tmp2->psu_address)
	    free(tmp2->psu_address);
	if (tmp2->profile_uuid)
	    free(tmp2->profile_uuid);
	if (tmp2->profile_name)
	    free(tmp2->profile_name);
	if (tmp2->profile_steps) {
	    for (step = tmp2->profile_steps; step; step = oldstep) {
		if (step->name)
		    free(step->name);
		oldstep = step->next;
		free(step);
	    }
	}
	if (tmp2->PID_cool)
	    free(tmp2->PID_cool);
	if (tmp2->PID_heat)
	    free(tmp2->PID_heat);
	free(tmp2);
    }
    Config.units = NULL;

    for (device = Config.devices; device; device = olddev) {
	olddev = device->next;
	if (device->uuid)
	    free(device->uuid);
	if (device->address)
	    free(device->address);
	if (device->description)
	    free(device->description);
	if (device->comment)
	    free(device->comment);
	free(device);
    }
    Config.devices = NULL;

#ifdef USE_SIMULATOR
    for (simulator = Config.simulators; simulator; simulator = oldsim) {
	oldsim = simulator->next;
	if (simulator->uuid)
	    free(simulator->uuid);
	if (simulator->name)
	    free(simulator->name);
	if (simulator->room_tempaddress)
	    free(simulator->room_tempaddress);
	if (simulator->room_humaddress)
	    free(simulator->room_humaddress);
	if (simulator->air_address)
	    free(simulator->air_address);
	if (simulator->beer_address)
	    free(simulator->beer_address);
	if (simulator->beer_address2)
	    free(simulator->beer_address2);
	if (simulator->chiller_address)
	    free(simulator->chiller_address);
	if (simulator->cooler_address)
	    free(simulator->cooler_address);
	if (simulator->heater_address)
	    free(simulator->heater_address);
	if (simulator->fan_address)
	    free(simulator->fan_address);
	if (simulator->light_address)
	    free(simulator->light_address);
	if (simulator->door_address)
	    free(simulator->door_address);
	if (simulator->psu_address)
	    free(simulator->psu_address);
	free(simulator);
    }
    Config.simulators = NULL;
#endif

    Config.lcd_cols = 16;
    Config.lcd_rows = 2;
}



/*
 * Write JSON.
 */
int do_wrconfig(void);
int do_wrconfig(void)
{
    char		*mypath = NULL;
    units_list		*tmp3;
    prof_step		*tmp5;
    devices_list	*device;
    json_object		*root, *jobj1;
#ifdef USE_SIMULATOR
    simulator_list	*simulator;
#endif

    root = json_object_new_object();
    jobj1 = json_object_new_object();
    json_object_object_add(root, "thermferm", jobj1);
    json_object_object_add(jobj1, "version", json_object_new_string(VERSION));
    json_object_object_add(jobj1, "name", json_object_new_string(Config.name));
    json_object_object_add(jobj1, "uuid", json_object_new_string(Config.uuid));
    json_object *port = json_object_new_object();
    json_object_object_add(port, "server", json_object_new_int(Config.server_port));
    json_object_object_add(port, "websocket", json_object_new_int(Config.websocket_port));
    json_object_object_add(jobj1, "port", port);

    json_object *temperature = json_object_new_object();
    if (Config.temp_uuid)
    	json_object_object_add(temperature, "uuid", json_object_new_string(Config.temp_uuid));
    else
	json_object_object_add(temperature, "uuid", NULL);
    /*
     * There is json_object_new_null() but that doesn't exist yet on version 0.12
     * which is on Slackware 14.2 Raspberries. Using NULL is exactly the same.
     */
    json_object_object_add(temperature, "state", json_object_new_string(DEVPRESENT[Config.temp_state]));
    json_object_object_add(temperature, "value", json_object_new_int(Config.temp_value));
    json_object_object_add(jobj1, "temperature", temperature);
    json_object *humidity = json_object_new_object();
    if (Config.hum_uuid)
	json_object_object_add(humidity, "uuid", json_object_new_string(Config.hum_uuid));
    else
	json_object_object_add(humidity, "uuid", NULL);
    json_object_object_add(humidity, "state", json_object_new_string(DEVPRESENT[Config.hum_state]));
    json_object_object_add(humidity, "value", json_object_new_int(Config.hum_value));
    json_object_object_add(jobj1, "humidity", humidity);
    json_object_object_add(jobj1, "temp_hum_idx", json_object_new_int(Config.temp_hum_idx));
    json_object_object_add(jobj1, "next_unit", json_object_new_int(Config.next_unit));

    json_object *mqtt = json_object_new_object();
    json_object_object_add(mqtt, "host", json_object_new_string(Config.mqtt_host));
    json_object_object_add(mqtt, "port", json_object_new_int(Config.mqtt_port));
    if (Config.mqtt_username && Config.mqtt_password) {
	json_object_object_add(mqtt, "username", json_object_new_string(Config.mqtt_username));
	json_object_object_add(mqtt, "password", json_object_new_string(Config.mqtt_password));
    }
    json_object_object_add(jobj1, "mqtt", mqtt);

    /*
     * LCD's. Maximum 7 but we only use one.
     */
    json_object *lcd = json_object_new_object();
    json_object_object_add(lcd, "address", json_object_new_int(Config.lcd_address));
    json_object_object_add(lcd, "columns", json_object_new_int(Config.lcd_cols));
    json_object_object_add(lcd, "rows", json_object_new_int(Config.lcd_rows));
    json_object_object_add(jobj1, "lcd", lcd);

    /*
     * Fermenter units
     */
    if (Config.units) {
	json_object *fermenters = json_object_new_array();
        json_object_object_add(jobj1, "fermenter", fermenters);

	for (tmp3 = Config.units; tmp3; tmp3 = tmp3->next) {
	    json_object *ferm = json_object_new_object();
            json_object_object_add(ferm, "uuid", json_object_new_string(tmp3->uuid));
	    json_object_object_add(ferm, "alias", json_object_new_string(tmp3->alias));
	    json_object_object_add(ferm, "mode", json_object_new_string(UNITMODE[tmp3->mode]));
	    json_object_object_add(ferm, "stage", json_object_new_string(UNITSTAGE[tmp3->stage]));

	    json_object *product = json_object_new_object();
	    json_object_object_add(product, "uuid", json_object_new_string(tmp3->product_uuid));
	    json_object_object_add(product, "code", json_object_new_string(tmp3->product_code));
	    json_object_object_add(product, "name", json_object_new_string(tmp3->product_name));
	    json_object_object_add(ferm, "product", product);

	    if (tmp3->air_address) {
		json_object *air = json_object_new_object();
		json_object_object_add(air, "address", json_object_new_string(tmp3->air_address));
		json_object_object_add(air, "state", json_object_new_string(DEVPRESENT[tmp3->air_state]));
		json_object_object_add(air, "temperature", json_object_new_int(tmp3->air_temperature));
		json_object_object_add(air, "idx", json_object_new_int(tmp3->air_idx));
		json_object_object_add(ferm, "air", air);
	    }
	    if (tmp3->beer_address) {
		json_object *beer = json_object_new_object();
                json_object_object_add(beer, "address", json_object_new_string(tmp3->beer_address));
                json_object_object_add(beer, "state", json_object_new_string(DEVPRESENT[tmp3->beer_state]));
                json_object_object_add(beer, "temperature", json_object_new_int(tmp3->beer_temperature));
                json_object_object_add(beer, "idx", json_object_new_int(tmp3->beer_idx));
		if (tmp3->beer_address2)
		    json_object_object_add(beer, "address2", json_object_new_string(tmp3->beer_address2));
                json_object_object_add(ferm, "beer", beer);
	    }
	    if (tmp3->chiller_address) {
		json_object *chiller = json_object_new_object();
                json_object_object_add(chiller, "address", json_object_new_string(tmp3->chiller_address));
                json_object_object_add(chiller, "state", json_object_new_string(DEVPRESENT[tmp3->chiller_state]));
                json_object_object_add(chiller, "temperature", json_object_new_int(tmp3->chiller_temperature));
                json_object_object_add(chiller, "idx", json_object_new_int(tmp3->chiller_idx));
                json_object_object_add(ferm, "chiller", chiller);
	    }
	    if (tmp3->heater_address) {
		json_object *heater = json_object_new_object();
                json_object_object_add(heater, "address", json_object_new_string(tmp3->heater_address));
                json_object_object_add(heater, "state", json_object_new_int(tmp3->heater_state));
                json_object_object_add(heater, "delay", json_object_new_int(tmp3->heater_delay));
		json_object_object_add(heater, "usage", json_object_new_int(tmp3->heater_usage));
                json_object_object_add(heater, "idx", json_object_new_int(tmp3->heater_idx));
                json_object_object_add(ferm, "heater", heater);
	    }
	    if (tmp3->cooler_address) {
		json_object *cooler = json_object_new_object();
                json_object_object_add(cooler, "address", json_object_new_string(tmp3->cooler_address));
                json_object_object_add(cooler, "state", json_object_new_int(tmp3->cooler_state));
                json_object_object_add(cooler, "delay", json_object_new_int(tmp3->cooler_delay));
                json_object_object_add(cooler, "usage", json_object_new_int(tmp3->cooler_usage));
                json_object_object_add(cooler, "idx", json_object_new_int(tmp3->cooler_idx));
                json_object_object_add(ferm, "cooler", cooler);
	    }
	    if (tmp3->fan_address) {
		json_object *fan = json_object_new_object();
                json_object_object_add(fan, "address", json_object_new_string(tmp3->fan_address));
                json_object_object_add(fan, "state", json_object_new_int(tmp3->fan_state));
                json_object_object_add(fan, "delay", json_object_new_int(tmp3->fan_delay));
                json_object_object_add(fan, "usage", json_object_new_int(tmp3->fan_usage));
                json_object_object_add(fan, "idx", json_object_new_int(tmp3->fan_idx));
                json_object_object_add(ferm, "fan", fan);
	    }
	    if (tmp3->light_address) {
		json_object *light = json_object_new_object();
                json_object_object_add(light, "address", json_object_new_string(tmp3->light_address));
                json_object_object_add(light, "state", json_object_new_int(tmp3->light_state));
                json_object_object_add(light, "delay", json_object_new_int(tmp3->light_delay));
                json_object_object_add(light, "usage", json_object_new_int(tmp3->light_usage));
                json_object_object_add(light, "idx", json_object_new_int(tmp3->light_idx));
                json_object_object_add(ferm, "light", light);
	    }
	    if (tmp3->door_address) {
		json_object *door = json_object_new_object();
                json_object_object_add(door, "address", json_object_new_string(tmp3->door_address));
                json_object_object_add(door, "state", json_object_new_int(tmp3->door_state));
                json_object_object_add(door, "idx", json_object_new_int(tmp3->door_idx));
                json_object_object_add(ferm, "door", door);
	    }
	    if (tmp3->psu_address) {
		json_object *psu = json_object_new_object();
                json_object_object_add(psu, "address", json_object_new_string(tmp3->psu_address));
                json_object_object_add(psu, "state", json_object_new_int(tmp3->psu_state));
                json_object_object_add(psu, "idx", json_object_new_int(tmp3->psu_idx));
                json_object_object_add(ferm, "psu", psu);
	    }
	    json_object *beer_set = json_object_new_object();
            json_object_object_add(beer_set, "low", json_object_new_double_fmt(tmp3->beer_set_lo, "%0.1f"));
            json_object_object_add(beer_set, "high", json_object_new_double_fmt(tmp3->beer_set_hi, "%0.1f"));
            json_object_object_add(ferm, "beer_set", beer_set);

	    json_object *fridge_set = json_object_new_object();
            json_object_object_add(fridge_set, "low", json_object_new_double_fmt(tmp3->fridge_set_lo, "%0.1f"));
            json_object_object_add(fridge_set, "high", json_object_new_double_fmt(tmp3->fridge_set_hi, "%0.1f"));
            json_object_object_add(ferm, "fridge_set", fridge_set);

	    json_object *temperature_set = json_object_new_object();
            json_object_object_add(temperature_set, "min", json_object_new_double_fmt(tmp3->temp_set_min, "%0.1f"));
            json_object_object_add(temperature_set, "max", json_object_new_double_fmt(tmp3->temp_set_max, "%0.1f"));
            json_object_object_add(ferm, "temperature_set", temperature_set);

	    json_object *yeast = json_object_new_object();
            json_object_object_add(yeast, "low", json_object_new_double_fmt(tmp3->yeast_lo, "%0.1f"));
            json_object_object_add(yeast, "high", json_object_new_double_fmt(tmp3->yeast_hi, "%0.1f"));
            json_object_object_add(ferm, "yeast", yeast);

	    if (tmp3->profile_uuid) {
		json_object *profile = json_object_new_object();
		json_object_object_add(profile, "uuid", json_object_new_string(tmp3->profile_uuid));
		json_object_object_add(profile, "name", json_object_new_string(tmp3->profile_name));
		json_object *inittemp = json_object_new_object();
		json_object_object_add(inittemp, "low", json_object_new_double_fmt(tmp3->profile_inittemp_lo, "%0.1f"));
		json_object_object_add(inittemp, "high", json_object_new_double_fmt(tmp3->profile_inittemp_hi, "%0.1f"));
		json_object_object_add(profile, "inittemp", inittemp);
		json_object_object_add(profile, "fridge_mode", json_object_new_int(tmp3->profile_fridge_mode));
		json_object_object_add(profile, "duration", json_object_new_int(tmp3->profile_duration));
		json_object_object_add(profile, "totalsteps", json_object_new_int(tmp3->profile_totalsteps));

		if (tmp3->profile_steps) {
		    json_object *psteps = json_object_new_array();
		    json_object_object_add(profile, "steps", psteps);

		    for (tmp5 = tmp3->profile_steps; tmp5; tmp5 = tmp5->next) {
			json_object *step = json_object_new_object();
			json_object_object_add(step, "resttime", json_object_new_int(tmp5->resttime));
			json_object_object_add(step, "steptime", json_object_new_int(tmp5->steptime));
			json_object_object_add(step, "target_lo", json_object_new_double_fmt(tmp5->target_lo, "%0.1f"));
			json_object_object_add(step, "target_hi", json_object_new_double_fmt(tmp5->target_hi, "%0.1f"));
			json_object_object_add(step, "fridge_mode", json_object_new_int(tmp5->fridge_mode));
			if (tmp5->name)
			    json_object_object_add(step, "name", json_object_new_string(tmp5->name));
			json_object_array_add(psteps, step);
		    }
		}

		json_object_object_add(profile, "started", json_object_new_int(tmp3->prof_started));
		json_object_object_add(profile, "paused", json_object_new_int(tmp3->prof_paused));
		json_object_object_add(profile, "state", json_object_new_string(PROFSTATE[tmp3->prof_state]));
		json_object_object_add(profile, "peak_abs", json_object_new_double_fmt(tmp3->prof_peak_abs, "%0.3f"));
		json_object_object_add(profile, "peak_rel", json_object_new_double_fmt(tmp3->prof_peak_rel, "%0.3f"));
		json_object_object_add(profile, "primary_done", json_object_new_int(tmp3->prof_primary_done));
		json_object_object_add(ferm, "profile", profile);
	    }
	    if (tmp3->PID_cool) {
		json_object *pidc = json_object_new_object();
		json_object_object_add(pidc, "iMax", json_object_new_double_fmt(tmp3->PID_cool->iMax, "%0.2f"));
		json_object_object_add(pidc, "iGain", json_object_new_double_fmt(tmp3->PID_cool->iGain, "%0.3f"));
		json_object_object_add(pidc, "pGain", json_object_new_double_fmt(tmp3->PID_cool->pGain, "%0.3f"));
		json_object_object_add(pidc, "dGain", json_object_new_double_fmt(tmp3->PID_cool->dGain, "%0.3f"));
		json_object_object_add(pidc, "idleRange", json_object_new_double_fmt(tmp3->PID_cool->idleRange, "%0.2f"));
		json_object_object_add(pidc, "Input", json_object_new_double_fmt(tmp3->PID_cool->Input, "%0.2f"));
		json_object_object_add(pidc, "Error", json_object_new_double_fmt(tmp3->PID_cool->Err, "%0.2f"));
		json_object_object_add(pidc, "iState", json_object_new_double_fmt(tmp3->PID_cool->iState, "%0.2f"));
		json_object_object_add(pidc, "SetP", json_object_new_double_fmt(tmp3->PID_cool->SetP, "%0.2f"));
		json_object_object_add(pidc, "OutP", json_object_new_double_fmt(tmp3->PID_cool->OutP, "%0.2f"));
		json_object_object_add(pidc, "Mode", json_object_new_string(PIDMODE[tmp3->PID_cool->Mode]));
		json_object_object_add(pidc, "Type", json_object_new_string("COOL"));
		json_object_object_add(ferm, "PIDcool", pidc);
	    }
	    if (tmp3->PID_heat) {
		json_object *pidh = json_object_new_object();
                json_object_object_add(pidh, "iMax", json_object_new_double_fmt(tmp3->PID_heat->iMax, "%0.2f"));
                json_object_object_add(pidh, "iGain", json_object_new_double_fmt(tmp3->PID_heat->iGain, "%0.3f"));
                json_object_object_add(pidh, "pGain", json_object_new_double_fmt(tmp3->PID_heat->pGain, "%0.3f"));
                json_object_object_add(pidh, "dGain", json_object_new_double_fmt(tmp3->PID_heat->dGain, "%0.3f"));
                json_object_object_add(pidh, "idleRange", json_object_new_double_fmt(tmp3->PID_heat->idleRange, "%0.2f"));
                json_object_object_add(pidh, "Input", json_object_new_double_fmt(tmp3->PID_heat->Input, "%0.2f"));
                json_object_object_add(pidh, "Error", json_object_new_double_fmt(tmp3->PID_heat->Err, "%0.2f"));
                json_object_object_add(pidh, "iState", json_object_new_double_fmt(tmp3->PID_heat->iState, "%0.2f"));
                json_object_object_add(pidh, "SetP", json_object_new_double_fmt(tmp3->PID_heat->SetP, "%0.2f"));
                json_object_object_add(pidh, "OutP", json_object_new_double_fmt(tmp3->PID_heat->OutP, "%0.2f"));
                json_object_object_add(pidh, "Mode", json_object_new_string(PIDMODE[tmp3->PID_heat->Mode]));
                json_object_object_add(pidh, "Type", json_object_new_string("HEAT"));
                json_object_object_add(ferm, "PIDheat", pidh);
	    }
	    json_object_array_add(fermenters, ferm);
	}
    }

    if (Config.devices) {
	json_object *devices = json_object_new_array();
        json_object_object_add(jobj1, "device", devices);

	pthread_mutex_lock(&mutexes[LOCK_DEVICES]);
	for (device = Config.devices; device; device = device->next) {
	    json_object *dev = json_object_new_object();
            json_object_object_add(dev, "uuid", json_object_new_string(device->uuid));
	    json_object_object_add(dev, "type", json_object_new_string(DEVTYPE[device->type]));
	    json_object_object_add(dev, "direction", json_object_new_string(DEVDIR[device->direction]));
	    json_object_object_add(dev, "value", json_object_new_int(device->value));
	    json_object_object_add(dev, "offset", json_object_new_int(device->offset));
	    json_object_object_add(dev, "present", json_object_new_string(DEVPRESENT[device->present]));
	    json_object_object_add(dev, "address", json_object_new_string(device->address));
	    json_object_object_add(dev, "subdevice", json_object_new_int(device->subdevice));
	    json_object_object_add(dev, "gpiopin", json_object_new_int(device->gpiopin));
	    json_object_object_add(dev, "description", json_object_new_string(device->description));
	    json_object_object_add(dev, "inuse", json_object_new_int(device->inuse));
	    json_object_object_add(dev, "comment", json_object_new_string(device->comment));
	    json_object_object_add(dev, "timestamp", json_object_new_int(device->timestamp));
	    json_object_array_add(devices, dev);
	}
	pthread_mutex_unlock(&mutexes[LOCK_DEVICES]);
    }

#ifdef USE_SIMULATOR
    if (Config.simulators) {

	json_object *simulators = json_object_new_array();
	json_object_object_add(jobj1, "simulator", simulators);

	for (simulator = Config.simulators; simulator; simulator = simulator->next) {
	    json_object *simu = json_object_new_object();
	    json_object_object_add(simu, "uuid", json_object_new_string(simulator->uuid));
	    json_object_object_add(simu, "name", json_object_new_string(simulator->name));
	    json_object_object_add(simu, "simno", json_object_new_int(simulator->simno));
	    json_object_object_add(simu, "volume_air", json_object_new_int(simulator->volume_air));
	    json_object_object_add(simu, "volume_beer", json_object_new_int(simulator->volume_beer));
	    json_object *room = json_object_new_object();
	    json_object_object_add(room, "tempaddress", json_object_new_string(simulator->room_tempaddress));
	    json_object_object_add(room, "temperature", json_object_new_double_fmt(simulator->room_temperature, "%0.1f"));
	    json_object_object_add(room, "humaddress", json_object_new_string(simulator->room_humaddress));
	    json_object_object_add(room, "humidity", json_object_new_double_fmt(simulator->room_humidity, "%0.1f"));
	    json_object_object_add(simu, "room", room);
	    json_object *air = json_object_new_object();
	    json_object_object_add(air, "address", json_object_new_string(simulator->air_address));
	    json_object_object_add(air, "temperature", json_object_new_double(simulator->air_temperature));
	    json_object_object_add(air, "present", json_object_new_string(DEVPRESENT[simulator->air_present]));
	    json_object_object_add(simu, "air", air);
	    json_object *beer = json_object_new_object();
            json_object_object_add(beer, "address", json_object_new_string(simulator->beer_address));
            json_object_object_add(beer, "temperature", json_object_new_double(simulator->beer_temperature));
            json_object_object_add(beer, "present", json_object_new_string(DEVPRESENT[simulator->beer_present]));
            json_object_object_add(simu, "beer", beer);
	    json_object *beer2 = json_object_new_object();
            json_object_object_add(beer2, "address", json_object_new_string(simulator->beer_address2));
            json_object_object_add(beer2, "temperature", json_object_new_double(simulator->beer_temperature2));
            json_object_object_add(beer2, "present", json_object_new_string(DEVPRESENT[simulator->beer_present2]));
            json_object_object_add(simu, "beer2", beer2);
	    json_object *chiller = json_object_new_object();
            json_object_object_add(chiller, "address", json_object_new_string(simulator->chiller_address));
            json_object_object_add(chiller, "temperature", json_object_new_double(simulator->chiller_temperature));
            json_object_object_add(chiller, "present", json_object_new_string(DEVPRESENT[simulator->chiller_present]));
            json_object_object_add(simu, "chiller", chiller);
	    json_object *cooler = json_object_new_object();
            json_object_object_add(cooler, "address", json_object_new_string(simulator->cooler_address));
            json_object_object_add(cooler, "temperature", json_object_new_double(simulator->cooler_temp));
	    json_object_object_add(cooler, "time", json_object_new_int(simulator->cooler_time));
	    json_object_object_add(cooler, "size", json_object_new_double_fmt(simulator->cooler_size, "%0.3f"));
            json_object_object_add(cooler, "present", json_object_new_string(DEVPRESENT[simulator->cooler_present]));
	    json_object_object_add(cooler, "power", json_object_new_int(simulator->cooler_power));
            json_object_object_add(simu, "cooler", cooler);
	    json_object *heater = json_object_new_object();
            json_object_object_add(heater, "address", json_object_new_string(simulator->heater_address));
            json_object_object_add(heater, "temperature", json_object_new_double(simulator->heater_temp));
            json_object_object_add(heater, "time", json_object_new_int(simulator->heater_time));
            json_object_object_add(heater, "size", json_object_new_double_fmt(simulator->heater_size, "%0.3f"));
            json_object_object_add(heater, "present", json_object_new_string(DEVPRESENT[simulator->heater_present]));
	    json_object_object_add(heater, "power", json_object_new_int(simulator->heater_power));
            json_object_object_add(simu, "heater", heater);
	    json_object *fan = json_object_new_object();
            json_object_object_add(fan, "address", json_object_new_string(simulator->fan_address));
            json_object_object_add(fan, "present", json_object_new_string(DEVPRESENT[simulator->fan_present]));
            json_object_object_add(fan, "power", json_object_new_int(simulator->fan_power));
            json_object_object_add(simu, "fan", fan);
	    json_object *light = json_object_new_object();
            json_object_object_add(light, "address", json_object_new_string(simulator->light_address));
            json_object_object_add(light, "present", json_object_new_string(DEVPRESENT[simulator->light_present]));
            json_object_object_add(light, "power", json_object_new_int(simulator->light_power));
            json_object_object_add(simu, "light", light);
	    json_object *door = json_object_new_object();
            json_object_object_add(door, "address", json_object_new_string(simulator->door_address));
            json_object_object_add(door, "present", json_object_new_string(DEVPRESENT[simulator->door_present]));
            json_object_object_add(door, "value", json_object_new_int(simulator->door_value));
            json_object_object_add(simu, "door", door);
	    json_object *psu = json_object_new_object();
            json_object_object_add(psu, "address", json_object_new_string(simulator->psu_address));
            json_object_object_add(psu, "present", json_object_new_string(DEVPRESENT[simulator->psu_present]));
            json_object_object_add(psu, "value", json_object_new_int(simulator->psu_value));
            json_object_object_add(simu, "psu", psu);
	    json_object_object_add(simu, "frigo_isolation", json_object_new_double(simulator->frigo_isolation));
	    json_object_object_add(simu, "timestamp", json_object_new_int(simulator->timestamp));
	    json_object_object_add(simu, "s_yeast_heat", json_object_new_double(simulator->s_yeast_heat));
	    json_object_object_add(simu, "s_yeast_started", json_object_new_int(simulator->s_yeast_started));
	    json_object_object_add(simu, "s_cool_temp", json_object_new_double(simulator->s_cool_temp));
	    json_object_object_add(simu, "s_heat_temp", json_object_new_double(simulator->s_heat_temp));
	    json_object_object_add(simu, "s_cool_changed", json_object_new_int(simulator->s_cool_changed));
	    json_object_object_add(simu, "s_heat_changed", json_object_new_int(simulator->s_heat_changed));
	    json_object_array_add(simulators, simu);
	}
    }
#endif

    /*
     * Now write the JSON configuration
     */
    if (getenv((char *)"USER") == NULL) {
        mypath = xstrcpy((char *)"/root");
    } else {
        mypath = xstrcpy(getenv((char *)"HOME"));
    }
    mypath = xstrcat(mypath, (char *)"/.config/thermferm/");
    mkdirs(mypath, 0755);
    mypath = xstrcat(mypath, (char *)"thermferm.json");

    if (debug)
        syslog(LOG_NOTICE, "Writing %s", mypath);

    if (json_object_to_file(mypath, root))
	syslog(LOG_NOTICE, "Error: failed to save %s", mypath);

    free(mypath);
    json_object_put(root);

    return 0;
}


int wrconfig(void)
{
    int rc = do_wrconfig();
    syslog(LOG_NOTICE, "Rewritten configuration, rc=%d", rc);
    return rc;
}


int present2int(const char *state)
{
    for (int i = 0; i < 4; i++)
	if (strcmp(state, DEVPRESENT[i]) == 0)
	    return i;
    return 0;
}


int mode2int(const char *state)
{
    for (int i = 0; i < 5; i++)
	if (strcmp(state, UNITMODE[i]) == 0)
	    return i;
    return 0;
}


int stage2int(const char *state)
{
    for (int i = 0; i < 4; i++)
        if (strcmp(state, UNITSTAGE[i]) == 0)
            return i;
    return 0;
}


int pidmode2int(const char *state)
{
    for (int i = 0; i < 3; i++)
        if (strcmp(state, PIDMODE[i]) == 0)
            return i;
    return 0;
}


int profstate2int(const char *state)
{
    for (int i = 0; i < 5; i++)
        if (strcmp(state, PROFSTATE[i]) == 0)
            return i;
    return 0;
}


int devtype2int(const char *state)
{
    for (int i = 0; i < 9; i++)
        if (strcmp(state, DEVTYPE[i]) == 0)
            return i;
    return 0;
}


int devpresent2int(const char *state)
{
    for (int i = 0; i < 4; i++)
        if (strcmp(state, DEVPRESENT[i]) == 0)
            return i;
    return 0;
}


int devdir2int(const char *state)
{
    for (int i = 0; i < 7; i++)
        if (strcmp(state, DEVDIR[i]) == 0)
            return i;
    return 0;
}


int rdconfig(void) 
{
    int			rc = 0;
    char		*mypath;
    uuid_t		uu;
    struct json_object	*val, *jobj0, *jobj1, *jobj2, *jobj3, *jobj4, *jobj5, *jobj6;
    units_list		*unit, *utmp;
    prof_step		*step, *ptmp;
    devices_list	*device, *dtmp;
#ifdef USE_SIMULATOR
    simulator_list	*simulator, *stmp;
#endif

    killconfig();
    syslog(LOG_NOTICE, "HOME='%s' USER='%s' LOGNAME='%s'", MBSE_SS(getenv((char *)"HOME")), MBSE_SS(getenv((char *)"USER")), MBSE_SS(getenv((char *)"LOGNAME")));

    /*
     * Search config file
     */
    if (getenv((char *)"USER") == NULL) {
	mypath = xstrcpy((char *)"/root");
    } else {
    	mypath = xstrcpy(getenv((char *)"HOME"));
    }
    mypath = xstrcat(mypath, (char *)"/.config/thermferm/");
    mkdirs(mypath, 0755);
    mypath = xstrcat(mypath, (char *)"thermferm.json");

    if (file_exist(mypath, F_OK)) {
	/*
	 * No config file, create a fresh one
	 */
	syslog(LOG_NOTICE, "rdconfig: %s not found, creating", mypath);
	Config.name = xstrcpy((char *)"NameMe");
	Config.uuid = malloc(37);
        uuid_generate(uu);
        uuid_unparse(uu, Config.uuid);
	wrconfig();
    }
    syslog(LOG_NOTICE, "rdconfig: using %s", mypath);

    Config.mqtt_host = xstrcpy((char *)"localhost");
    Config.mqtt_port = 1883;
    Config.next_unit = 1;
    Config.websocket_port = 8020;

    jobj0 = json_object_from_file(mypath);
    if (! jobj0) {
	syslog(LOG_NOTICE, "rdconfig: could not open %s", mypath);
	free(mypath);
	return 1;
    }

    if (! json_object_object_get_ex(jobj0, "thermferm", &jobj1)) {
	syslog(LOG_NOTICE, "%s not thermferm json config", mypath);
	free(mypath);
	return 1;
    }
    free(mypath);
    mypath = NULL;

    if (json_object_object_get_ex(jobj1, "name", &val))
	Config.name = xstrcpy((char *)json_object_get_string(val));
    if (json_object_object_get_ex(jobj1, "uuid", &val))
	Config.uuid = xstrcpy((char *)json_object_get_string(val));
    if (json_object_object_get_ex(jobj1, "temp_hum_idx", &val))
	Config.temp_hum_idx = json_object_get_int(val);
    if (json_object_object_get_ex(jobj1, "next_unit", &val))
	Config.next_unit = json_object_get_int(val);
    if (json_object_object_get_ex(jobj1, "port", &jobj2)) {
	if (json_object_object_get_ex(jobj2, "server", &val))
	    Config.server_port = json_object_get_int(val);
	if (json_object_object_get_ex(jobj2, "websocket", &val))
	    Config.websocket_port = json_object_get_int(val);
    }
    if (json_object_object_get_ex(jobj1, "temperature", &jobj2)) {
	if (json_object_object_get_ex(jobj2, "uuid", &val))
	    Config.temp_uuid = xstrcpy((char *)json_object_get_string(val));
	// Restore state and value ?
    }
    if (json_object_object_get_ex(jobj1, "humidity", &jobj2)) {
	if (json_object_object_get_ex(jobj2, "uuid", &val))
	    Config.hum_uuid = xstrcpy((char *)json_object_get_string(val));
	// Restore state and value ?
    }
    if (json_object_object_get_ex(jobj1, "mqtt", &jobj2)) {
	if (json_object_object_get_ex(jobj2, "host", &val)) {
	    if (Config.mqtt_host)
		free(Config.mqtt_host);
	    Config.mqtt_host = xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(jobj2, "port", &val))
	    Config.mqtt_port = json_object_get_int(val);
	if (json_object_object_get_ex(jobj2, "username", &val))
	    Config.mqtt_username = xstrcpy((char *)json_object_get_string(val));
	if (json_object_object_get_ex(jobj2, "password", &val))
	    Config.mqtt_password = xstrcpy((char *)json_object_get_string(val));
    }
    if (json_object_object_get_ex(jobj1, "lcd", &jobj2)) {
	if (json_object_object_get_ex(jobj2, "address", &val))
	    Config.lcd_address = json_object_get_int(val);
	if (json_object_object_get_ex(jobj2, "columns", &val))
	    Config.lcd_cols = json_object_get_int(val);
	if (json_object_object_get_ex(jobj2, "rows", &val))
	    Config.lcd_rows = json_object_get_int(val);
    }

    /*
     * Parse fermenters array
     */
    if (json_object_object_get_ex(jobj1, "fermenter", &jobj2)) {
	for (int j = 0; j < json_object_array_length(jobj2); j++) {
	    /*
	     * Process one fermenter unit
	     */
	    jobj3 = json_object_array_get_idx(jobj2, j);
	    unit = (units_list *)malloc(sizeof(units_list));
	    unit->next = NULL;
	    unit->uuid = unit->product_uuid = unit->product_code = unit->product_name = unit->event_msg = \
	    unit->alias = unit->air_address = unit->beer_address = unit->beer_address2 = unit->chiller_address = \
			  unit->heater_address = unit->cooler_address = unit->fan_address = unit->door_address = \
			  unit->light_address = unit->psu_address = unit->profile_uuid = unit->profile_name  = NULL;
	    unit->prof_peak_abs = unit->prof_peak_rel = 0.0;
	    unit->air_temperature = unit->beer_temperature = unit->chiller_temperature = unit->beer_set_lo = unit->beer_set_hi = \
			  unit->fridge_set_lo = unit->fridge_set_hi = unit->profile_inittemp_lo = unit->profile_inittemp_hi = 20.0;
	    unit->air_state = unit->beer_state = unit->chiller_state = DEVPRESENT_NO; // missing
	    unit->heater_state = unit->cooler_state = unit->fan_state = unit->door_state = \
	    unit->light_state = unit->light_timer = unit->psu_state = unit->mode = unit->prof_state = unit->stage = 0;
	    unit->air_idx = unit->beer_idx = unit->chiller_idx = unit->heater_idx = unit->cooler_idx = unit->fan_idx = \
			  unit->door_idx = unit->light_idx = unit->psu_idx = unit->profile_fridge_mode = \
			  unit->profile_duration = unit->profile_totalsteps = 0;
	    unit->profile_steps = NULL;
	    unit->heater_delay = unit->cooler_delay = unit->fan_delay = 20;     /* 5 minutes delay */
	    unit->light_delay = 1;                                              /* 15 seconds delay     */
	    unit->heater_wait = unit->cooler_wait = unit->fan_wait = unit->light_wait = 0;
	    unit->heater_usage = unit->cooler_usage = unit->fan_usage = unit->light_usage = 0;
	    unit->temp_set_min = 1.0;
	    unit->temp_set_max = 45.0;
	    unit->yeast_lo = 12.0;
	    unit->yeast_hi = 24.0;
	    unit->prof_started = unit->prof_paused = unit->prof_primary_done = (time_t)0;
	    unit->prof_percent = 0;
	    unit->PID_cool = (pid_var *)malloc(sizeof(pid_var));
	    unit->PID_heat = (pid_var *)malloc(sizeof(pid_var));
	    InitPID(unit->PID_cool, PID_TYPE_COOL);
	    InitPID(unit->PID_heat, PID_TYPE_HEAT);
	    unit->product_uuid = malloc(37);
            uuid_generate(uu);
            uuid_unparse(uu, unit->product_uuid);
	    unit->product_code = xstrcpy((char *)"FAKE0000");
	    unit->product_name = xstrcpy((char *)"No beer");

	    if (json_object_object_get_ex(jobj3, "uuid", &val))
               	unit->uuid = xstrcpy((char *)json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "alias", &val))
		unit->alias = xstrcpy((char *)json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "mode", &val))
		unit->mode = mode2int(json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "stage", &val))
		unit->stage = stage2int(json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "product", &jobj4)) {
		if (json_object_object_get_ex(jobj4, "uuid", &val)) {
		    if (unit->product_uuid)
			free(unit->product_uuid);
		    unit->product_uuid = xstrcpy((char *)json_object_get_string(val));
		}
		if (json_object_object_get_ex(jobj4, "code", &val)) {
		    if (unit->product_code)
			free(unit->product_code);
		    unit->product_code = xstrcpy((char *)json_object_get_string(val));
		}
		if (json_object_object_get_ex(jobj4, "name", &val)) {
		    if (unit->product_name)
			free(unit->product_name);
		    unit->product_name = xstrcpy((char *)json_object_get_string(val));
		}
	    }
	    if (json_object_object_get_ex(jobj3, "air", &jobj4)) {
		if (json_object_object_get_ex(jobj4, "address", &val))
		    unit->air_address = xstrcpy((char *)json_object_get_string(val));
		if (json_object_object_get_ex(jobj4, "state", &val))
		    unit->air_state = present2int(json_object_get_string(val));
		if (json_object_object_get_ex(jobj4, "idx", &val))
		    unit->air_idx = json_object_get_int(val);
		if (json_object_object_get_ex(jobj4, "temperature", &val))
		    unit->air_temperature = json_object_get_int(val);
	    }
	    if (json_object_object_get_ex(jobj3, "beer", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    unit->beer_address = xstrcpy((char *)json_object_get_string(val));
		if (json_object_object_get_ex(jobj4, "address2", &val))
                    unit->beer_address2 = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "state", &val))
                    unit->beer_state = present2int(json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "idx", &val))
                    unit->beer_idx = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "temperature", &val))
                    unit->beer_temperature = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "chiller", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    unit->chiller_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "state", &val))
                    unit->chiller_state = present2int(json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "idx", &val))
                    unit->chiller_idx = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "temperature", &val))
                    unit->chiller_temperature = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "heater", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    unit->heater_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "idx", &val))
                    unit->heater_idx = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "delay", &val))
                    unit->heater_delay = json_object_get_int(val);
		if (json_object_object_get_ex(jobj4, "usage", &val))
                    unit->heater_usage = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "cooler", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    unit->cooler_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "idx", &val))
                    unit->cooler_idx = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "delay", &val))
                    unit->cooler_delay = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "usage", &val))
                    unit->cooler_usage = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "fan", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    unit->fan_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "idx", &val))
                    unit->fan_idx = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "delay", &val))
                    unit->fan_delay = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "usage", &val))
                    unit->fan_usage = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "light", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    unit->light_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "idx", &val))
                    unit->light_idx = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "delay", &val))
                    unit->light_delay = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "usage", &val))
                    unit->light_usage = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "door", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    unit->door_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "idx", &val))
                    unit->door_idx = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "psu", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    unit->psu_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "idx", &val))
                    unit->psu_idx = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "beer_set", &jobj4)) {
		if (json_object_object_get_ex(jobj4, "low", &val))
		    unit->beer_set_lo = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "high", &val))
                    unit->beer_set_hi = json_object_get_double(val);
	    }
	    if (json_object_object_get_ex(jobj3, "fridge_set", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "low", &val))
                    unit->fridge_set_lo = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "high", &val))
                    unit->fridge_set_hi = json_object_get_double(val);
            }
	    if (json_object_object_get_ex(jobj3, "temperature_set", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "min", &val))
                    unit->temp_set_min = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "max", &val))
                    unit->temp_set_max = json_object_get_double(val);
            }
	    if (json_object_object_get_ex(jobj3, "yeast", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "low", &val))
                    unit->yeast_lo = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "high", &val))
                    unit->yeast_hi = json_object_get_double(val);
            }
	    if (json_object_object_get_ex(jobj3, "PIDcool", &jobj4)) {
		if (json_object_object_get_ex(jobj4, "iMax", &val))
		    unit->PID_cool->iMax = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "idleRange", &val))
		    unit->PID_cool->idleRange = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "iGain", &val))
		    unit->PID_cool->iGain = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "pGain", &val))
                    unit->PID_cool->pGain = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "dGain", &val))
                    unit->PID_cool->dGain = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "Input", &val))
                    unit->PID_cool->Input = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "Error", &val))
                    unit->PID_cool->Err = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "iState", &val))
                    unit->PID_cool->iState = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "SetP", &val))
                    unit->PID_cool->SetP = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "OutP", &val))
                    unit->PID_cool->OutP = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "Mode", &val))
		    unit->PID_cool->Mode = pidmode2int(json_object_get_string(val));
	    }
	    if (json_object_object_get_ex(jobj3, "PIDheat", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "iMax", &val))
                    unit->PID_heat->iMax = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "idleRange", &val))
                    unit->PID_heat->idleRange = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "iGain", &val))
                    unit->PID_heat->iGain = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "pGain", &val))
                    unit->PID_heat->pGain = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "dGain", &val))
                    unit->PID_heat->dGain = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "Input", &val))
                    unit->PID_heat->Input = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "Error", &val))
                    unit->PID_heat->Err = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "iState", &val))
                    unit->PID_heat->iState = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "SetP", &val))
                    unit->PID_heat->SetP = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "OutP", &val))
                    unit->PID_heat->OutP = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "Mode", &val))
                    unit->PID_heat->Mode = pidmode2int(json_object_get_string(val));
            }
	    if (json_object_object_get_ex(jobj3, "profile", &jobj4)) {
		if (json_object_object_get_ex(jobj4, "uuid", &val))
		    unit->profile_uuid = xstrcpy((char *)json_object_get_string(val));
		if (json_object_object_get_ex(jobj4, "name", &val))
		    unit->profile_name = xstrcpy((char *)json_object_get_string(val));
		if (json_object_object_get_ex(jobj4, "inittemp", &jobj5)) {
		    if (json_object_object_get_ex(jobj5, "low", &val))
                        unit->profile_inittemp_lo = json_object_get_double(val);
                    if (json_object_object_get_ex(jobj5, "high", &val))
                        unit->profile_inittemp_hi = json_object_get_double(val);
		}
		if (json_object_object_get_ex(jobj4, "fridge_mode", &val))
		    unit->profile_fridge_mode = json_object_get_int(val);
		if (json_object_object_get_ex(jobj4, "duration", &val))
                    unit->profile_duration = json_object_get_int(val);
		if (json_object_object_get_ex(jobj4, "totalsteps", &val))
                    unit->profile_totalsteps = json_object_get_int(val);
		if (json_object_object_get_ex(jobj4, "started", &val))
                    unit->prof_started = json_object_get_int(val);
		if (json_object_object_get_ex(jobj4, "paused", &val))
                    unit->prof_paused = json_object_get_int(val);
		if (json_object_object_get_ex(jobj4, "peak_abs", &val))
		    unit->prof_peak_abs = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "peak_rel", &val))
                    unit->prof_peak_rel = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "primary_done", &val))
                    unit->prof_primary_done = json_object_get_int(val);
		if (json_object_object_get_ex(jobj4, "state", &val))
		    unit->prof_state = profstate2int(json_object_get_string(val));

		if (json_object_object_get_ex(jobj4, "steps", &jobj5)) {
		    for (int k = 0; k < json_object_array_length(jobj5); k++) {
			jobj6 = json_object_array_get_idx(jobj5, k);
			step = (prof_step *)malloc(sizeof(prof_step));
			step->next = NULL;
			step->name = NULL;
			step->steptime = step->resttime = step->fridge_mode = 0;
			step->target_lo = step->target_hi = 20.0;

			if (json_object_object_get_ex(jobj6, "name", &val))
			    step->name = xstrcpy((char *)json_object_get_string(val));
			if (json_object_object_get_ex(jobj6, "resttime", &val))
			    step->resttime = json_object_get_int(val);
			if (json_object_object_get_ex(jobj6, "steptime", &val))
                            step->steptime = json_object_get_int(val);
			if (json_object_object_get_ex(jobj6, "target_lo", &val))
			    step->target_lo = json_object_get_double(val);
			if (json_object_object_get_ex(jobj6, "target_hi", &val))
                            step->target_hi = json_object_get_double(val);
			if (json_object_object_get_ex(jobj6, "fridge_mode", &val))
                            step->fridge_mode = json_object_get_int(val);

			if (unit->profile_steps == NULL) {
			    unit->profile_steps = step;
			} else {
			    for (ptmp = unit->profile_steps; ptmp; ptmp = ptmp->next) {
				if (ptmp->next == NULL) {
				    ptmp->next = step;
				    break;
				}
			    }
			}
		    }
		}
	    }

	    if (Config.units == NULL) {
		Config.units = unit;
	    } else {
		for (utmp = Config.units; utmp; utmp = utmp->next) {
		    if (utmp->next == NULL) {
			utmp->next = unit;
			break;
		    }
		}
	    }
	}
    } /* fermenters array */

    /*
     * Parse devices array
     */
    if (json_object_object_get_ex(jobj1, "device", &jobj2)) {
	for (int j = 0; j < json_object_array_length(jobj2); j++) {
	    /*
	     * Process one device
	     */
	    jobj3 = json_object_array_get_idx(jobj2, j);
	    device = (devices_list *)malloc(sizeof(devices_list));
	    device->next = NULL;
	    device->uuid = device->address = device->description = device->comment = NULL;
	    device->type = device->direction = device->present = device->subdevice = device->inuse = device->offset = 0;
	    device->gpiopin = -1;
	    device->timestamp = (time_t)0;

	    if (json_object_object_get_ex(jobj3, "uuid", &val))
		device->uuid = xstrcpy((char *)json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "type", &val))
		device->type = devtype2int(json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "direction", &val))
		device->direction = devdir2int(json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "value", &val))
		device->value = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "offset", &val))
		device->offset = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "present", &val))
		device->present = devpresent2int(json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "address", &val))
		device->address = xstrcpy((char *)json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "subdevice", &val))
		device->subdevice = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "gpiopin", &val))
		device->gpiopin = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "description", &val))
		device->description = xstrcpy((char *)json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "inuse", &val))
		device->inuse = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "comment", &val))
		device->comment = xstrcpy((char *)json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "timestamp", &val))
		device->timestamp = json_object_get_int(val);


	    if (Config.devices == NULL) {
		Config.devices = device;
	    } else {
		for (dtmp = Config.devices; dtmp; dtmp = dtmp->next) {
		    if (dtmp->next == NULL) {
			dtmp->next = device;
			break;
		    }
		}
	    }
	}
    } /* devices array */

#ifdef USE_SIMULATOR
    if (json_object_object_get_ex(jobj1, "simulator", &jobj2)) {
	for (int j = 0; j < json_object_array_length(jobj2); j++) {
            /*
             * Process one simulator
             */
            jobj3 = json_object_array_get_idx(jobj2, j);
	    simulator = (simulator_list *)malloc(sizeof(simulator_list));
	    simulator->next = NULL;
	    simulator->uuid = simulator->name = NULL;
	    simulator->room_tempaddress = simulator->room_humaddress = simulator->air_address = simulator->beer_address = simulator->beer_address2 = NULL;
	    simulator->chiller_address = simulator->cooler_address = simulator->heater_address = simulator->fan_address = simulator->light_address = NULL;
	    simulator->door_address = simulator->psu_address = NULL;
	    simulator->simno = simulator->volume_air = simulator->volume_beer = 0;
	    simulator->room_temperature = simulator->air_temperature = simulator->beer_temperature = simulator->s_cool_temp = simulator->s_heat_temp = 20.0;
	    simulator->chiller_temperature = 1.5;
	    simulator->room_humidity = 49.2;
	    simulator->cooler_temp = simulator->cooler_size = simulator->heater_temp = simulator->heater_size = simulator->frigo_isolation = 0.0;
	    simulator->cooler_time = simulator->heater_time = 0;
	    simulator->air_present = simulator->beer_present = DEVPRESENT_YES;
	    simulator->beer_present2 = simulator->chiller_present = simulator->cooler_present = simulator->heater_present = DEVPRESENT_UNDEF;
	    simulator->s_yeast_started = simulator->s_cool_changed = simulator->s_heat_changed = (time_t)0;
	    simulator->s_yeast_heat = simulator->s_cool_temp = simulator->s_heat_temp = 0.0;

	    if (json_object_object_get_ex(jobj3, "uuid", &val))
		simulator->uuid = xstrcpy((char *)json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "name", &val))
		simulator->name = xstrcpy((char *)json_object_get_string(val));
	    if (json_object_object_get_ex(jobj3, "simno", &val))
		simulator->simno = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "volume_air", &val))
		simulator->volume_air = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "volume_beer", &val))
		simulator->volume_beer = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "room", &jobj4)) {
		if (json_object_object_get_ex(jobj4, "tempaddress", &val))
		    simulator->room_tempaddress = xstrcpy((char *)json_object_get_string(val));
		if (json_object_object_get_ex(jobj4, "temperature", &val))
		    simulator->room_temperature = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "humaddress", &val))
		    simulator->room_humaddress = xstrcpy((char *)json_object_get_string(val));
		if (json_object_object_get_ex(jobj4, "humidity", &val))
		    simulator->room_humidity = json_object_get_double(val);
	    }
	    if (json_object_object_get_ex(jobj3, "air", &jobj4)) {
		if (json_object_object_get_ex(jobj4, "address", &val))
		    simulator->air_address = xstrcpy((char *)json_object_get_string(val));
		if (json_object_object_get_ex(jobj4, "temperature", &val))
		    simulator->air_temperature = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "present", &val))
		    simulator->air_present = present2int(json_object_get_string(val));
	    }
	    if (json_object_object_get_ex(jobj3, "beer", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    simulator->beer_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "temperature", &val))
                    simulator->beer_temperature = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "present", &val))
                    simulator->beer_present = present2int(json_object_get_string(val));
            }
	    if (json_object_object_get_ex(jobj3, "beer", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    simulator->beer_address2 = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "temperature", &val))
                    simulator->beer_temperature2 = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "present", &val))
                    simulator->beer_present2 = present2int(json_object_get_string(val));
            }
	    if (json_object_object_get_ex(jobj3, "chiller", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    simulator->chiller_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "temperature", &val))
                    simulator->chiller_temperature = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "present", &val))
                    simulator->chiller_present = present2int(json_object_get_string(val));
            }
	    if (json_object_object_get_ex(jobj3, "cooler", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    simulator->cooler_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "temperature", &val))
                    simulator->cooler_temp = json_object_get_double(val);
		if (json_object_object_get_ex(jobj4, "time", &val))
		    simulator->cooler_time = json_object_get_int(val);
		if (json_object_object_get_ex(jobj4, "size", &val))
		    simulator->cooler_size = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "present", &val))
                    simulator->cooler_present = present2int(json_object_get_string(val));
		if (json_object_object_get_ex(jobj4, "power", &val))
		    simulator->cooler_power = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "heater", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    simulator->heater_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "temperature", &val))
                    simulator->heater_temp = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "time", &val))
                    simulator->heater_time = json_object_get_int(val);
                if (json_object_object_get_ex(jobj4, "size", &val))
                    simulator->heater_size = json_object_get_double(val);
                if (json_object_object_get_ex(jobj4, "present", &val))
                    simulator->heater_present = present2int(json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "power", &val))
                    simulator->heater_power = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "fan", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    simulator->fan_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "present", &val))
                    simulator->fan_present = present2int(json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "power", &val))
                    simulator->fan_power = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "light", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    simulator->light_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "present", &val))
                    simulator->light_present = present2int(json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "power", &val))
                    simulator->light_power = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "door", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    simulator->door_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "present", &val))
                    simulator->door_present = present2int(json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "value", &val))
                    simulator->door_value = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "psu", &jobj4)) {
                if (json_object_object_get_ex(jobj4, "address", &val))
                    simulator->psu_address = xstrcpy((char *)json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "present", &val))
                    simulator->psu_present = present2int(json_object_get_string(val));
                if (json_object_object_get_ex(jobj4, "value", &val))
                    simulator->psu_value = json_object_get_int(val);
            }
	    if (json_object_object_get_ex(jobj3, "frigo_isolation", &val))
                simulator->frigo_isolation = json_object_get_double(val);
	    if (json_object_object_get_ex(jobj3, "timestamp", &val))
                simulator->timestamp = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "s_yeast_heat", &val))
		simulator->s_yeast_heat = json_object_get_double(val);
	    if (json_object_object_get_ex(jobj3, "s_yeast_started", &val))
		simulator->s_yeast_started = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "s_cool_temp", &val))
		simulator->s_cool_temp = json_object_get_double(val);
	    if (json_object_object_get_ex(jobj3, "s_heat_temp", &val))
		simulator->s_heat_temp = json_object_get_double(val);
	    if (json_object_object_get_ex(jobj3, "s_cool_changed", &val))
		simulator->s_cool_changed = json_object_get_int(val);
	    if (json_object_object_get_ex(jobj3, "s_heat_changed", &val))
		simulator->s_heat_changed = json_object_get_int(val);

	    if (Config.simulators == NULL) {
	        Config.simulators = simulator;
	    } else {
	        for (stmp = Config.simulators; stmp; stmp = stmp->next) {
	            if (stmp->next == NULL) {
	                stmp->next = simulator;
	                break;
	            }
	        }
	    }
	}
    } /* simulators array */
#endif

    /*
     * If the system uuid is not set, do it now.
     */
    if (Config.uuid && ! strcmp((char *)"(null)", Config.uuid)) {
	free(Config.uuid);
	Config.uuid = NULL;
    }
    if (Config.uuid == NULL) {
	Config.uuid = malloc(37);
	uuid_generate(uu);
	uuid_unparse(uu, Config.uuid);
    }

    return rc;
}



